#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------------------------------------------
# INFO:
# -----------------------------------------------------------------------------------------------------------------------

"""
Authors: Evan Hubinger, Fred Buchanan, Noah Lipsyc, Ishaan Verma
License: Apache 2.0
Description: The Coconut command-line utility.
"""

# -----------------------------------------------------------------------------------------------------------------------
# IMPORTS:
# -----------------------------------------------------------------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division

from coconut.root import *  # NOQA

import sys
import os
import time
import shutil
from contextlib import contextmanager
from subprocess import CalledProcessError

from coconut._pyparsing import (
    unset_fast_pyparsing_reprs,
    collect_timing_info,
    print_timing_info,
)

from coconut.compiler import Compiler
from coconut.exceptions import (
    CoconutException,
    CoconutInternalException,
)
from coconut.terminal import (
    logger,
    format_error,
    internal_assert,
)
from coconut.constants import (
    PY32,
    PY35,
    fixpath,
    code_exts,
    comp_ext,
    watch_interval,
    icoconut_default_kernel_names,
    icoconut_default_kernel_dirs,
    icoconut_custom_kernel_name,
    icoconut_old_kernel_names,
    exit_chars,
    coconut_run_args,
    coconut_run_verbose_args,
    verbose_mypy_args,
    default_mypy_args,
    report_this_text,
    mypy_silent_non_err_prefixes,
    mypy_silent_err_prefixes,
    mypy_err_infixes,
    mypy_install_arg,
    mypy_builtin_regex,
    coconut_pth_file,
    error_color_code,
)
from coconut.util import (
    univ_open,
    ver_tuple_to_str,
    install_custom_kernel,
    get_clock_time,
    first_import_time,
)
from coconut.command.util import (
    writefile,
    readfile,
    showpath,
    rem_encoding,
    Runner,
    multiprocess_wrapper,
    Prompt,
    handling_broken_process_pool,
    kill_children,
    run_cmd,
    set_mypy_path,
    is_special_dir,
    launch_documentation,
    launch_tutorial,
    stdin_readable,
    set_recursion_limit,
    can_parse,
    invert_mypy_arg,
)
from coconut.compiler.util import (
    should_indent,
    get_target_info_smart,
)
from coconut.compiler.header import gethash
from coconut.compiler.grammar import set_grammar_names
from coconut.command.cli import arguments, cli_version

# -----------------------------------------------------------------------------------------------------------------------
# MAIN:
# -----------------------------------------------------------------------------------------------------------------------


class Command(object):
    """Coconut command-line interface."""
    comp = None  # current coconut.compiler.Compiler
    show = False  # corresponds to --display flag
    runner = None  # the current Runner
    jobs = 0  # corresponds to --jobs flag
    executor = None  # runs --jobs
    exit_code = 0  # exit status to return
    errmsg = None  # error message to display
    mypy_args = None  # corresponds to --mypy flag
    argv_args = None  # corresponds to --argv flag

    def __init__(self):
        """Create the CLI."""
        self.prompt = Prompt()

    def start(self, run=False):
        """Endpoint for coconut and coconut-run."""
        if run:
            args, argv = [], []
            # for coconut-run, all args beyond the source file should be wrapped in an --argv
            for i in range(1, len(sys.argv)):
                arg = sys.argv[i]
                args.append(arg)
                # if arg is source file, put everything else in argv
                if not arg.startswith("-") and can_parse(arguments, args[:-1]):
                    argv = sys.argv[i + 1:]
                    break
            for run_arg in (coconut_run_verbose_args if "--verbose" in args else coconut_run_args):
                if run_arg not in args:
                    args.append(run_arg)
            self.cmd(args, argv=argv)
        else:
            self.cmd()

    def cmd(self, args=None, argv=None, interact=True, default_target=None):
        """Process command-line arguments."""
        if args is None:
            parsed_args = arguments.parse_args()
        else:
            parsed_args = arguments.parse_args(args)
        if argv is not None:
            if parsed_args.argv is not None:
                raise CoconutException("cannot pass --argv/--args when using coconut-run (coconut-run interprets any arguments after the source file as --argv/--args)")
            parsed_args.argv = argv
        if parsed_args.target is None:
            parsed_args.target = default_target
        self.exit_code = 0
        with self.handling_exceptions():
            self.use_args(parsed_args, interact, original_args=args)
        self.exit_on_error()

    def setup(self, *args, **kwargs):
        """Set parameters for the compiler."""
        if self.comp is None:
            self.comp = Compiler(*args, **kwargs)
        else:
            self.comp.setup(*args, **kwargs)

    def parse_block(self, code):
        """Compile a block of code for the interpreter."""
        return self.comp.parse_block(code, keep_state=True)

    def exit_on_error(self):
        """Exit if exit_code is abnormal."""
        if self.exit_code:
            if self.errmsg is not None:
                # show on stdout with error color code so that stdout
                #  listeners see the error
                logger.show("Coconut exiting with error: " + self.errmsg, color=error_color_code)
                self.errmsg = None
            if self.using_jobs:
                kill_children()
            sys.exit(self.exit_code)

    def use_args(self, args, interact=True, original_args=None):
        """Handle command-line arguments."""
        # fix args
        if not DEVELOP:
            args.trace = args.profile = False

        # set up logger
        logger.quiet, logger.verbose, logger.tracing = args.quiet, args.verbose, args.trace
        if args.verbose or args.trace or args.profile:
            set_grammar_names()
        if args.trace or args.profile:
            unset_fast_pyparsing_reprs()
        if args.profile:
            collect_timing_info()
        logger.enable_colors()

        logger.log(cli_version)
        if original_args is not None:
            logger.log("Directly passed args:", original_args)
        logger.log("Parsed args:", args)

        # validate general command args
        if args.mypy is not None and args.line_numbers:
            logger.warn("extraneous --line-numbers argument passed; --mypy implies --line-numbers")
        if args.site_install and args.site_uninstall:
            raise CoconutException("cannot --site-install and --site-uninstall simultaneously")
        for and_args in getattr(args, "and") or []:
            if len(and_args) > 2:
                raise CoconutException(
                    "--and accepts at most two arguments, source and dest ({n} given: {args!r})".format(
                        n=len(and_args),
                        args=and_args,
                    ),
                )

        # process general command args
        if args.recursion_limit is not None:
            set_recursion_limit(args.recursion_limit)
        if args.jobs is not None:
            self.set_jobs(args.jobs)
        if args.display:
            self.show = True
        if args.style is not None:
            self.prompt.set_style(args.style)
        if args.history_file is not None:
            self.prompt.set_history_file(args.history_file)
        if args.vi_mode:
            self.prompt.vi_mode = True
        if args.docs:
            launch_documentation()
        if args.tutorial:
            launch_tutorial()
        if args.site_uninstall:
            self.site_uninstall()
        if args.site_install:
            self.site_install()
        if args.argv is not None:
            self.argv_args = list(args.argv)

        # additional validation after processing
        if args.profile and self.jobs != 0:
            raise CoconutException("--profile incompatible with --jobs {jobs}".format(jobs=args.jobs))

        # process general compiler args
        self.setup(
            target=args.target,
            strict=args.strict,
            minify=args.minify,
            line_numbers=args.line_numbers or args.mypy is not None,
            keep_lines=args.keep_lines,
            no_tco=args.no_tco,
            no_wrap=args.no_wrap,
        )

        # process mypy args and print timing info (must come after compiler setup)
        if args.mypy is not None:
            self.set_mypy_args(args.mypy)
        logger.log("Grammar init time: " + str(self.comp.grammar_init_time) + " secs / Total init time: " + str(get_clock_time() - first_import_time) + " secs")

        if args.source is not None:
            # warnings if source is given
            if args.interact and args.run:
                logger.warn("extraneous --run argument passed; --interact implies --run")
            if args.package and self.mypy:
                logger.warn("extraneous --package argument passed; --mypy implies --package")

            # errors if source is given
            if args.standalone and args.package:
                raise CoconutException("cannot compile as both --package and --standalone")
            if args.standalone and self.mypy:
                raise CoconutException("cannot compile as both --package (implied by --mypy) and --standalone")
            if args.no_write and self.mypy:
                raise CoconutException("cannot compile with --no-write when using --mypy")

            # process all source, dest pairs
            src_dest_package_triples = [
                self.process_source_dest(src, dst, args)
                for src, dst in (
                    [(args.source, args.dest)]
                    + (getattr(args, "and") or [])
                )
            ]

            # do compilation
            with self.running_jobs(exit_on_error=not args.watch):
                filepaths = []
                for source, dest, package in src_dest_package_triples:
                    filepaths += self.compile_path(source, dest, package, run=args.run or args.interact, force=args.force)
            self.run_mypy(filepaths)

        # validate args if no source is given
        elif (
            args.run
            or args.no_write
            or args.force
            or args.package
            or args.standalone
            or args.watch
        ):
            raise CoconutException("a source file/folder must be specified when options that depend on the source are enabled")
        elif getattr(args, "and"):
            raise CoconutException("--and should only be used for extra source/dest pairs, not the first source/dest pair")

        # handle extra cli tasks
        if args.code is not None:
            self.execute(self.parse_block(args.code))
        got_stdin = False
        if args.jupyter is not None:
            self.start_jupyter(args.jupyter)
        elif stdin_readable():
            logger.log("Reading piped input from stdin...")
            self.execute(self.parse_block(sys.stdin.read()))
            got_stdin = True
        if args.interact or (
            interact and not (
                got_stdin
                or args.source
                or args.code
                or args.tutorial
                or args.docs
                or args.watch
                or args.site_uninstall
                or args.site_install
                or args.jupyter is not None
                or args.mypy == [mypy_install_arg]
            )
        ):
            self.start_prompt()
        if args.watch:
            # src_dest_package_triples is always available here
            self.watch(src_dest_package_triples, args.run, args.force)
        if args.profile:
            print_timing_info()

    def process_source_dest(self, source, dest, args):
        """Determine the correct source, dest, package mode to use for the given source, dest, and args."""
        # determine source
        processed_source = fixpath(source)

        # validate args
        if (args.run or args.interact) and os.path.isdir(processed_source):
            if args.run:
                raise CoconutException("source path %r must point to file not directory when --run is enabled" % (source,))
            if args.interact:
                raise CoconutException("source path %r must point to file not directory when --run (implied by --interact) is enabled" % (source,))
        if args.watch and os.path.isfile(processed_source):
            raise CoconutException("source path %r must point to directory not file when --watch is enabled" % (source,))

        # determine dest
        if dest is None:
            if args.no_write:
                processed_dest = False  # no dest
            else:
                processed_dest = True  # auto-generate dest
        elif args.no_write:
            raise CoconutException("destination path cannot be given when --no-write is enabled")
        else:
            processed_dest = dest

        # determine package mode
        if args.package or self.mypy:
            package = True
        elif args.standalone:
            package = False
        else:
            # auto-decide package
            if os.path.isfile(source):
                package = False
            elif os.path.isdir(source):
                package = True
            else:
                raise CoconutException("could not find source path", source)

        return processed_source, processed_dest, package

    def register_exit_code(self, code=1, errmsg=None, err=None):
        """Update the exit code."""
        if err is not None:
            internal_assert(errmsg is None, "register_exit_code accepts only one of errmsg or err")
            if logger.verbose:
                errmsg = format_error(err)
            else:
                errmsg = err.__class__.__name__
        if errmsg is not None:
            if self.errmsg is None:
                self.errmsg = errmsg
            elif errmsg not in self.errmsg:
                if logger.verbose:
                    self.errmsg += "\nAnd error: " + errmsg
                else:
                    self.errmsg += "; " + errmsg
        if code is not None:
            self.exit_code = code or self.exit_code

    @contextmanager
    def handling_exceptions(self):
        """Perform proper exception handling."""
        try:
            if self.using_jobs:
                with handling_broken_process_pool():
                    yield
            else:
                yield
        except SystemExit as err:
            self.register_exit_code(err.code)
        except BaseException as err:
            if isinstance(err, CoconutException):
                logger.print_exc()
            elif not isinstance(err, KeyboardInterrupt):
                logger.print_exc()
                logger.printerr(report_this_text)
            self.register_exit_code(err=err)

    def compile_path(self, path, write=True, package=True, **kwargs):
        """Compile a path and returns paths to compiled files."""
        if not isinstance(write, bool):
            write = fixpath(write)
        if os.path.isfile(path):
            destpath = self.compile_file(path, write, package, **kwargs)
            return [destpath] if destpath is not None else []
        elif os.path.isdir(path):
            return self.compile_folder(path, write, package, **kwargs)
        else:
            raise CoconutException("could not find source path", path)

    def compile_folder(self, directory, write=True, package=True, **kwargs):
        """Compile a directory and returns paths to compiled files."""
        if not isinstance(write, bool) and os.path.isfile(write):
            raise CoconutException("destination path cannot point to a file when compiling a directory")
        filepaths = []
        for dirpath, dirnames, filenames in os.walk(directory):
            if isinstance(write, bool):
                writedir = write
            else:
                writedir = os.path.join(write, os.path.relpath(dirpath, directory))
            for filename in filenames:
                if os.path.splitext(filename)[1] in code_exts:
                    with self.handling_exceptions():
                        destpath = self.compile_file(os.path.join(dirpath, filename), writedir, package, **kwargs)
                        if destpath is not None:
                            filepaths.append(destpath)
            for name in dirnames[:]:
                if not is_special_dir(name) and name.startswith("."):
                    if logger.verbose:
                        logger.show_tabulated("Skipped directory", name, "(explicitly pass as source to override).")
                    dirnames.remove(name)  # directories removed from dirnames won't appear in further os.walk iterations
        return filepaths

    def compile_file(self, filepath, write=True, package=False, force=False, **kwargs):
        """Compile a file and returns the compiled file's path."""
        set_ext = False
        if write is False:
            destpath = None
        elif write is True:
            destpath = filepath
            set_ext = True
        elif os.path.splitext(write)[1]:
            # write is a file; it is the destination filepath
            destpath = write
        else:
            # write is a dir; make the destination filepath by adding the filename
            destpath = os.path.join(write, os.path.basename(filepath))
            set_ext = True
        if set_ext:
            base, ext = os.path.splitext(os.path.splitext(destpath)[0])
            if not ext:
                ext = comp_ext
            destpath = fixpath(base + ext)
        if filepath == destpath:
            raise CoconutException("cannot compile " + showpath(filepath) + " to itself", extra="incorrect file extension")
        if destpath is not None:
            dest_ext = os.path.splitext(destpath)[1]
            if dest_ext in code_exts:
                if force:
                    logger.warn("found destination path with " + dest_ext + " extension; compiling anyway due to --force")
                else:
                    raise CoconutException(
                        "found destination path with " + dest_ext + " extension; aborting compilation",
                        extra="pass --force to override",
                    )
        self.compile(filepath, destpath, package, force=force, **kwargs)
        return destpath

    def compile(self, codepath, destpath=None, package=False, run=False, force=False, show_unchanged=True):
        """Compile a source Coconut file to a destination Python file."""
        with univ_open(codepath, "r") as opened:
            code = readfile(opened)

        package_level = -1
        if destpath is not None:
            destpath = fixpath(destpath)
            destdir = os.path.dirname(destpath)
            if not os.path.exists(destdir):
                os.makedirs(destdir)
            if package is True:
                package_level = self.get_package_level(codepath)
                if package_level == 0:
                    self.create_package(destdir)

        foundhash = None if force else self.has_hash_of(destpath, code, package_level)
        if foundhash:
            if show_unchanged:
                logger.show_tabulated("Left unchanged", showpath(destpath), "(pass --force to override).")
            if self.show:
                logger.print(foundhash)
            if run:
                self.execute_file(destpath, argv_source_path=codepath)

        else:
            logger.show_tabulated("Compiling", showpath(codepath), "...")

            def callback(compiled):
                if destpath is None:
                    logger.show_tabulated("Compiled", showpath(codepath), "without writing to file.")
                else:
                    with univ_open(destpath, "w") as opened:
                        writefile(opened, compiled)
                    logger.show_tabulated("Compiled to", showpath(destpath), ".")
                if self.show:
                    logger.print(compiled)
                if run:
                    if destpath is None:
                        self.execute(compiled, path=codepath, allow_show=False)
                    else:
                        self.execute_file(destpath, argv_source_path=codepath)

            if package is True:
                self.submit_comp_job(codepath, callback, "parse_package", code, package_level=package_level)
            elif package is False:
                self.submit_comp_job(codepath, callback, "parse_file", code)
            else:
                raise CoconutInternalException("invalid value for package", package)

    def get_package_level(self, codepath):
        """Get the relative level to the base directory of the package."""
        package_level = -1
        check_dir = os.path.dirname(os.path.abspath(codepath))
        while check_dir:
            has_init = False
            for ext in code_exts:
                init_file = os.path.join(check_dir, "__init__" + ext)
                if os.path.exists(init_file):
                    has_init = True
                    break
            if has_init:
                package_level += 1
                check_dir = os.path.dirname(check_dir)
            else:
                break
        if package_level < 0:
            if self.comp.strict:
                logger.warn("missing __init__" + code_exts[0] + " in package", check_dir, extra="remove --strict to dismiss")
            package_level = 0
        return package_level
        return 0

    def create_package(self, dirpath):
        """Set up a package directory."""
        filepath = os.path.join(dirpath, "__coconut__.py")
        with univ_open(filepath, "w") as opened:
            writefile(opened, self.comp.getheader("__coconut__"))

    def submit_comp_job(self, path, callback, method, *args, **kwargs):
        """Submits a job on self.comp to be run in parallel."""
        if self.executor is None:
            with self.handling_exceptions():
                callback(getattr(self.comp, method)(*args, **kwargs))
        else:
            path = showpath(path)
            with logger.in_path(path):  # pickle the compiler in the path context
                future = self.executor.submit(multiprocess_wrapper(self.comp, method), *args, **kwargs)

            def callback_wrapper(completed_future):
                """Ensures that all errors are always caught, since errors raised in a callback won't be propagated."""
                with logger.in_path(path):  # handle errors in the path context
                    with self.handling_exceptions():
                        result = completed_future.result()
                        callback(result)
            future.add_done_callback(callback_wrapper)

    def set_jobs(self, jobs):
        """Set --jobs."""
        if jobs == "sys":
            self.jobs = None
        else:
            try:
                jobs = int(jobs)
            except ValueError:
                jobs = -1  # will raise error below
            if jobs < 0:
                raise CoconutException("--jobs must be an integer >= 0 or 'sys'")
            self.jobs = jobs

    @property
    def using_jobs(self):
        """Determine whether or not multiprocessing is being used."""
        return self.jobs is None or self.jobs > 1

    @contextmanager
    def running_jobs(self, exit_on_error=True):
        """Initialize multiprocessing."""
        with self.handling_exceptions():
            if self.using_jobs:
                from concurrent.futures import ProcessPoolExecutor
                try:
                    with ProcessPoolExecutor(self.jobs) as self.executor:
                        yield
                finally:
                    self.executor = None
            else:
                yield
        if exit_on_error:
            self.exit_on_error()

    def has_hash_of(self, destpath, code, package_level):
        """Determine if a file has the hash of the code."""
        if destpath is not None and os.path.isfile(destpath):
            with univ_open(destpath, "r") as opened:
                compiled = readfile(opened)
            hashash = gethash(compiled)
            if hashash is not None and hashash == self.comp.genhash(code, package_level):
                return True
        return False

    def get_input(self, more=False):
        """Prompt for code input."""
        received = None
        try:
            received = self.prompt.input(more)
        except KeyboardInterrupt:
            logger.printerr("\nKeyboardInterrupt")
        except EOFError:
            logger.print()
            self.exit_runner()
        else:
            if received.startswith(exit_chars):
                self.exit_runner()
                received = None
        return received

    def start_running(self):
        """Start running the Runner."""
        self.comp.warm_up()
        self.check_runner()
        self.running = True
        logger.log("Time till prompt: " + str(get_clock_time() - first_import_time) + " secs")

    def start_prompt(self):
        """Start the interpreter."""
        logger.show(
            "Coconut Interpreter v{co_ver}:".format(
                co_ver=VERSION,
            ),
        )
        logger.show("(enter 'exit()' or press Ctrl-D to end)")
        self.start_running()
        while self.running:
            try:
                code = self.get_input()
                if code:
                    compiled = self.handle_input(code)
                    if compiled:
                        self.execute(compiled, use_eval=None)
            except KeyboardInterrupt:
                logger.printerr("\nKeyboardInterrupt")

    def exit_runner(self, exit_code=0):
        """Exit the interpreter."""
        self.register_exit_code(exit_code)
        self.running = False

    def handle_input(self, code):
        """Compile Coconut interpreter input."""
        if not self.prompt.multiline:
            if not should_indent(code):
                try:
                    return self.parse_block(code)
                except CoconutException:
                    pass
            while True:
                line = self.get_input(more=True)
                if line is None:
                    return None
                elif line:
                    code += "\n" + line
                else:
                    break
        try:
            return self.parse_block(code)
        except CoconutException:
            logger.print_exc()
        return None

    def execute(self, compiled=None, path=None, use_eval=False, allow_show=True):
        """Execute compiled code."""
        self.check_runner()
        if compiled is not None:

            if allow_show and self.show:
                logger.print(compiled)

            if path is None:  # header is not included
                if not self.mypy:
                    no_str_code = self.comp.remove_strs(compiled)
                    result = mypy_builtin_regex.search(no_str_code)
                    if result:
                        logger.warn("found mypy-only built-in " + repr(result.group(0)) + "; pass --mypy to use mypy-only built-ins at the interpreter")

            else:  # header is included
                compiled = rem_encoding(compiled)

            self.runner.run(compiled, use_eval=use_eval, path=path, all_errors_exit=path is not None)

            self.run_mypy(code=self.runner.was_run_code())

    def execute_file(self, destpath, **kwargs):
        """Execute compiled file."""
        self.check_runner(**kwargs)
        self.runner.run_file(destpath)

    def check_runner(self, set_sys_vars=True, argv_source_path=""):
        """Make sure there is a runner."""
        if set_sys_vars:
            # set sys.path
            if os.getcwd() not in sys.path:
                sys.path.append(os.getcwd())

            # set sys.argv
            if self.argv_args is not None:
                sys.argv = [argv_source_path] + self.argv_args

        # set up runner
        if self.runner is None:
            self.runner = Runner(self.comp, exit=self.exit_runner, store=self.mypy)

        # pass runner to prompt
        self.prompt.set_runner(self.runner)

    @property
    def mypy(self):
        """Whether using MyPy or not."""
        return self.mypy_args is not None

    def set_mypy_args(self, mypy_args=None):
        """Set MyPy arguments."""
        if mypy_args is None:
            self.mypy_args = None

        elif mypy_install_arg in mypy_args:
            if mypy_args != [mypy_install_arg]:
                raise CoconutException("'--mypy install' cannot be used alongside other --mypy arguments")
            stub_dir = set_mypy_path()
            logger.show_sig("Successfully installed MyPy stubs into " + repr(stub_dir))
            self.mypy_args = None

        else:
            self.mypy_args = list(mypy_args)

            if not any(arg.startswith("--python-version") for arg in self.mypy_args):
                self.mypy_args += [
                    "--python-version",
                    ver_tuple_to_str(get_target_info_smart(self.comp.target, mode="highest")),
                ]

            if not any(arg.startswith("--python-executable") for arg in self.mypy_args):
                self.mypy_args += [
                    "--python-executable",
                    sys.executable,
                ]

            add_mypy_args = default_mypy_args + (verbose_mypy_args if logger.verbose else ())

            for arg in add_mypy_args:
                no_arg = invert_mypy_arg(arg)
                arg_prefixes = (arg,) + ((no_arg,) if no_arg is not None else ())
                if not any(arg.startswith(arg_prefixes) for arg in self.mypy_args):
                    self.mypy_args.append(arg)

            logger.log("MyPy args:", self.mypy_args)
            self.mypy_errs = []

    def run_mypy(self, paths=(), code=None):
        """Run MyPy with arguments."""
        if self.mypy:
            set_mypy_path()
            from coconut.command.mypy import mypy_run
            args = list(paths) + self.mypy_args
            if code is not None:  # interpreter
                args += ["-c", code]
            for line, is_err in mypy_run(args):
                line = line.rstrip()
                logger.log("[MyPy]", line)
                if line.startswith(mypy_silent_err_prefixes):
                    if code is None:  # file
                        logger.printerr(line)
                        self.register_exit_code(errmsg="MyPy error")
                elif not line.startswith(mypy_silent_non_err_prefixes):
                    if code is None:  # file
                        logger.printerr(line)
                        if any(infix in line for infix in mypy_err_infixes):
                            self.register_exit_code(errmsg="MyPy error")
                    if line not in self.mypy_errs:
                        if code is not None:  # interpreter
                            logger.printerr(line)
                        self.mypy_errs.append(line)

    def run_silent_cmd(self, *args):
        """Same as run_cmd$(show_output=logger.verbose)."""
        return run_cmd(*args, show_output=logger.verbose)

    def install_jupyter_kernel(self, jupyter, kernel_dir):
        """Install the given kernel via the command line and return whether successful."""
        install_args = jupyter + ["kernelspec", "install", kernel_dir, "--replace"]
        try:
            self.run_silent_cmd(install_args)
        except CalledProcessError:
            user_install_args = install_args + ["--user"]
            try:
                self.run_silent_cmd(user_install_args)
            except CalledProcessError:
                logger.warn("kernel install failed on command", " ".join(install_args))
                self.register_exit_code(errmsg="Jupyter kernel error")
                return False
        return True

    def remove_jupyter_kernel(self, jupyter, kernel_name):
        """Remove the given kernel via the command line and return whether successful."""
        remove_args = jupyter + ["kernelspec", "remove", kernel_name, "-f"]
        try:
            self.run_silent_cmd(remove_args)
        except CalledProcessError:
            logger.warn("kernel removal failed on command", " ".join(remove_args))
            self.register_exit_code(errmsg="Jupyter kernel error")
            return False
        return True

    def install_default_jupyter_kernels(self, jupyter, kernel_list):
        """Install icoconut default kernels."""
        logger.show_sig("Installing Jupyter kernels '" + "', '".join(icoconut_default_kernel_names) + "'...")
        overall_success = True

        for old_kernel_name in icoconut_old_kernel_names:
            if old_kernel_name in kernel_list:
                success = self.remove_jupyter_kernel(jupyter, old_kernel_name)
                overall_success = overall_success and success

        for kernel_dir in icoconut_default_kernel_dirs:
            success = self.install_jupyter_kernel(jupyter, kernel_dir)
            overall_success = overall_success and success

        if overall_success:
            return icoconut_default_kernel_names
        else:
            return []

    def get_jupyter_kernels(self, jupyter):
        """Get the currently installed Jupyter kernels."""
        raw_kernel_list = run_cmd(jupyter + ["kernelspec", "list"], show_output=False, raise_errs=False)

        kernel_list = []
        for line in raw_kernel_list.splitlines():
            kernel_list.append(line.split()[0])
        return kernel_list

    def get_jupyter_command(self):
        """Get the correct jupyter command."""
        for jupyter in (
            [sys.executable, "-m", "jupyter"],
            [sys.executable, "-m", "ipython"],
        ):
            if PY35:  # newer Python versions should only use "jupyter", not "ipython"
                break
            try:
                self.run_silent_cmd(jupyter + ["--help"])  # --help is much faster than --version
            except CalledProcessError:
                logger.warn("failed to find Jupyter command at " + repr(" ".join(jupyter)))
            else:
                break
        else:  # no break
            raise CoconutException("'coconut --jupyter' requires Jupyter (run 'pip install coconut[jupyter]' to fix)")
        return jupyter

    def start_jupyter(self, args):
        """Start Jupyter with the Coconut kernel."""
        jupyter = self.get_jupyter_command()

        # get a list of installed kernels
        kernel_list = self.get_jupyter_kernels(jupyter)
        newly_installed_kernels = []

        # always update the custom kernel, but only reinstall it if it isn't already there or given no args
        custom_kernel_dir = install_custom_kernel(logger=logger)
        if custom_kernel_dir is not None and (icoconut_custom_kernel_name not in kernel_list or not args):
            logger.show_sig("Installing Jupyter kernel {name!r}...".format(name=icoconut_custom_kernel_name))
            if self.install_jupyter_kernel(jupyter, custom_kernel_dir):
                newly_installed_kernels.append(icoconut_custom_kernel_name)

        if not args:
            # install default kernels if given no args
            newly_installed_kernels += self.install_default_jupyter_kernels(jupyter, kernel_list)
            run_args = None

        else:
            # use the custom kernel if it exists
            if icoconut_custom_kernel_name in kernel_list or icoconut_custom_kernel_name in newly_installed_kernels:
                kernel = icoconut_custom_kernel_name

            # otherwise determine which default kernel to use and install them if necessary
            else:
                ver = "2" if PY2 else "3"
                try:
                    self.run_silent_cmd(["python" + ver, "-m", "coconut.main", "--version"])
                except CalledProcessError:
                    kernel = "coconut_py"
                else:
                    kernel = "coconut_py" + ver
                if kernel not in kernel_list:
                    newly_installed_kernels += self.install_default_jupyter_kernels(jupyter, kernel_list)
                logger.warn("could not find {name!r} kernel; using {kernel!r} kernel instead".format(name=icoconut_custom_kernel_name, kernel=kernel))

            # pass the kernel to the console or otherwise just launch Jupyter now that we know our kernel is available
            if args[0] == "console":
                run_args = jupyter + ["console", "--kernel", kernel] + args[1:]
            else:
                run_args = jupyter + args

        if newly_installed_kernels:
            logger.show_sig("Successfully installed Jupyter kernels: '" + "', '".join(newly_installed_kernels) + "'")

        # run the Jupyter command
        if run_args is not None:
            self.register_exit_code(run_cmd(run_args, raise_errs=False), errmsg="Jupyter error")

    def watch(self, src_dest_package_triples, run=False, force=False):
        """Watch a source and recompile on change."""
        from coconut.command.watch import Observer, RecompilationWatcher

        for src, _, _ in src_dest_package_triples:
            logger.show()
            logger.show_tabulated("Watching", showpath(src), "(press Ctrl-C to end)...")

        def recompile(path, src, dest, package):
            path = fixpath(path)
            if os.path.isfile(path) and os.path.splitext(path)[1] in code_exts:
                with self.handling_exceptions():
                    if dest is True or dest is None:
                        writedir = dest
                    else:
                        # correct the compilation path based on the relative position of path to src
                        dirpath = os.path.dirname(path)
                        writedir = os.path.join(dest, os.path.relpath(dirpath, src))
                    filepaths = self.compile_path(path, writedir, package, run=run, force=force, show_unchanged=False)
                    self.run_mypy(filepaths)

        observer = Observer()
        watchers = []
        for src, dest, package in src_dest_package_triples:
            watcher = RecompilationWatcher(recompile, src, dest, package)
            observer.schedule(watcher, src, recursive=True)
            watchers.append(watcher)

        with self.running_jobs():
            observer.start()
            try:
                while True:
                    time.sleep(watch_interval)
                    for wcher in watchers:
                        wcher.keep_watching()
            except KeyboardInterrupt:
                logger.show_sig("Got KeyboardInterrupt; stopping watcher.")
            finally:
                observer.stop()
                observer.join()

    def get_python_lib(self):
        """Get current Python lib location."""
        # these are expensive, so should only be imported here
        if PY32:
            from sysconfig import get_path
            python_lib = get_path("purelib")
        else:
            from distutils import sysconfig
            python_lib = sysconfig.get_python_lib()
        return fixpath(python_lib)

    def site_install(self):
        """Add Coconut's pth file to site-packages."""
        python_lib = self.get_python_lib()

        shutil.copy(coconut_pth_file, python_lib)
        logger.show_sig("Added %s to %s" % (os.path.basename(coconut_pth_file), python_lib))

    def site_uninstall(self):
        """Remove Coconut's pth file from site-packages."""
        python_lib = self.get_python_lib()
        pth_file = os.path.join(python_lib, os.path.basename(coconut_pth_file))

        if os.path.isfile(pth_file):
            os.remove(pth_file)
            logger.show_sig("Removed %s from %s" % (os.path.basename(coconut_pth_file), python_lib))
        else:
            raise CoconutException("failed to find %s file to remove" % (os.path.basename(coconut_pth_file),))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------------------------------------------
# INFO:
# -----------------------------------------------------------------------------------------------------------------------

"""
Author: Evan Hubinger
License: Apache 2.0
Description: Convenience functions for using Coconut as a module.
"""

# -----------------------------------------------------------------------------------------------------------------------
# IMPORTS:
# -----------------------------------------------------------------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division

from coconut.root import *  # NOQA

import sys
import os.path
import codecs
try:
    from encodings import utf_8
except ImportError:
    utf_8 = None

from coconut.integrations import embed
from coconut.exceptions import CoconutException
from coconut.command import Command
from coconut.command.cli import cli_version
from coconut.compiler import Compiler
from coconut.constants import (
    version_tag,
    code_exts,
    coconut_import_hook_args,
    coconut_kernel_kwargs,
)

# -----------------------------------------------------------------------------------------------------------------------
# COMMAND:
# -----------------------------------------------------------------------------------------------------------------------

GLOBAL_STATE = None


def get_state(state=None):
    """Get a Coconut state object; None gets a new state, False gets the global state."""
    global GLOBAL_STATE
    if state is None:
        return Command()
    elif state is False:
        if GLOBAL_STATE is None:
            GLOBAL_STATE = Command()
        return GLOBAL_STATE
    else:
        return state


def cmd(cmd_args, interact=False, state=False, **kwargs):
    """Process command-line arguments."""
    if isinstance(cmd_args, (str, bytes)):
        cmd_args = cmd_args.split()
    return get_state(state).cmd(cmd_args, interact=interact, **kwargs)


VERSIONS = {
    "num": VERSION,
    "name": VERSION_NAME,
    "spec": VERSION_STR,
    "tag": version_tag,
    "-v": cli_version,
}


def version(which="num"):
    """Get the Coconut version."""
    if which in VERSIONS:
        return VERSIONS[which]
    else:
        raise CoconutException(
            "invalid version type " + repr(which),
            extra="valid versions are " + ", ".join(VERSIONS),
        )


# -----------------------------------------------------------------------------------------------------------------------
# COMPILER:
# -----------------------------------------------------------------------------------------------------------------------

def setup(*args, **kwargs):
    """Set up the given state object."""
    state = kwargs.pop("state", False)
    return get_state(state).setup(*args, **kwargs)


PARSERS = {
    "sys": lambda comp: comp.parse_sys,
    "exec": lambda comp: comp.parse_exec,
    "file": lambda comp: comp.parse_file,
    "package": lambda comp: comp.parse_package,
    "block": lambda comp: comp.parse_block,
    "single": lambda comp: comp.parse_single,
    "eval": lambda comp: comp.parse_eval,
    "lenient": lambda comp: comp.parse_lenient,
    "xonsh": lambda comp: comp.parse_xonsh,
}

# deprecated aliases
PARSERS["any"] = PARSERS["debug"] = PARSERS["lenient"]


def parse(code="", mode="sys", state=False, keep_internal_state=None):
    """Compile Coconut code."""
    if keep_internal_state is None:
        keep_internal_state = bool(state)
    command = get_state(state)
    if command.comp is None:
        command.setup()
    if mode not in PARSERS:
        raise CoconutException(
            "invalid parse mode " + repr(mode),
            extra="valid modes are " + ", ".join(PARSERS),
        )
    return PARSERS[mode](command.comp)(code, keep_state=keep_internal_state)


def coconut_eval(expression, globals=None, locals=None, state=False, **kwargs):
    """Compile and evaluate Coconut code."""
    command = get_state(state)
    if command.comp is None:
        setup()
    command.check_runner(set_sys_vars=False)
    if globals is None:
        globals = {}
    command.runner.update_vars(globals)
    compiled_python = parse(expression, "eval", state, **kwargs)
    return eval(compiled_python, globals, locals)


# -----------------------------------------------------------------------------------------------------------------------
# BREAKPOINT:
# -----------------------------------------------------------------------------------------------------------------------


def _coconut_breakpoint():
    """Determine coconut.embed depth based on whether we're being
    called by Coconut's breakpoint() or Python's breakpoint()."""
    if sys.version_info >= (3, 7):
        return embed(depth=1)
    else:
        return embed(depth=2)


def use_coconut_breakpoint(on=True):
    """Switches the breakpoint() built-in (universally accessible via
    coconut.__coconut__.breakpoint) to use coconut.embed."""
    if on:
        sys.breakpointhook = _coconut_breakpoint
    else:
        sys.breakpointhook = sys.__breakpointhook__


use_coconut_breakpoint()


# -----------------------------------------------------------------------------------------------------------------------
# AUTOMATIC COMPILATION:
# -----------------------------------------------------------------------------------------------------------------------


class CoconutImporter(object):
    """Finder and loader for compiling Coconut files at import time."""
    ext = code_exts[0]

    @staticmethod
    def run_compiler(path):
        """Run the Coconut compiler on the given path."""
        cmd([path] + list(coconut_import_hook_args))

    def find_module(self, fullname, path=None):
        """Searches for a Coconut file of the given name and compiles it."""
        basepaths = [""] + list(sys.path)
        if fullname.startswith("."):
            if path is None:
                # we can't do a relative import if there's no package path
                return
            fullname = fullname[1:]
            basepaths.insert(0, path)
        fullpath = os.path.join(*fullname.split("."))
        for head in basepaths:
            path = os.path.join(head, fullpath)
            filepath = path + self.ext
            dirpath = os.path.join(path, "__init__" + self.ext)
            if os.path.exists(filepath):
                self.run_compiler(filepath)
                # Coconut file was found and compiled, now let Python import it
                return
            if os.path.exists(dirpath):
                self.run_compiler(path)
                # Coconut package was found and compiled, now let Python import it
                return


coconut_importer = CoconutImporter()


def auto_compilation(on=True):
    """Turn automatic compilation of Coconut files on or off."""
    if on:
        if coconut_importer not in sys.meta_path:
            sys.meta_path.insert(0, coconut_importer)
    else:
        try:
            sys.meta_path.remove(coconut_importer)
        except ValueError:
            pass


auto_compilation()


# -----------------------------------------------------------------------------------------------------------------------
# ENCODING:
# -----------------------------------------------------------------------------------------------------------------------


if utf_8 is not None:
    class CoconutStreamReader(utf_8.StreamReader, object):
        """Compile Coconut code from a stream of UTF-8."""
        coconut_compiler = None

        @classmethod
        def compile_coconut(cls, source):
            """Compile the given Coconut source text."""
            if cls.coconut_compiler is None:
                cls.coconut_compiler = Compiler(**coconut_kernel_kwargs)
            return cls.coconut_compiler.parse_sys(source)

        @classmethod
        def decode(cls, input_bytes, errors="strict"):
            """Decode and compile the given Coconut source bytes."""
            input_str, len_consumed = super(CoconutStreamReader, cls).decode(input_bytes, errors)
            return cls.compile_coconut(input_str), len_consumed

    class CoconutIncrementalDecoder(utf_8.IncrementalDecoder, object):
        """Compile Coconut at the end of incrementally decoding UTF-8."""
        invertible = False
        _buffer_decode = CoconutStreamReader.decode


def get_coconut_encoding(encoding="coconut"):
    """Get a CodecInfo for the given Coconut encoding."""
    if not encoding.startswith("coconut"):
        return None
    if encoding != "coconut":
        raise CoconutException("unknown Coconut encoding: " + repr(encoding))
    if utf_8 is None:
        raise CoconutException("coconut encoding requires encodings.utf_8")
    return codecs.CodecInfo(
        name=encoding,
        encode=utf_8.encode,
        decode=CoconutStreamReader.decode,
        incrementalencoder=utf_8.IncrementalEncoder,
        incrementaldecoder=CoconutIncrementalDecoder,
        streamreader=CoconutStreamReader,
        streamwriter=utf_8.StreamWriter,
    )


codecs.register(get_coconut_encoding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------------------------------------------
# INFO:
# -----------------------------------------------------------------------------------------------------------------------

"""
Author: Evan Hubinger
License: Apache 2.0
Description: Handles Coconut pattern-matching.
"""

# -----------------------------------------------------------------------------------------------------------------------
# IMPORTS:
# -----------------------------------------------------------------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division

from coconut.root import *  # NOQA

from contextlib import contextmanager
from collections import OrderedDict

from coconut._pyparsing import ParseResults
from coconut.terminal import (
    internal_assert,
    logger,
)
from coconut.exceptions import (
    CoconutInternalException,
    CoconutDeferredSyntaxError,
    CoconutSyntaxWarning,
)
from coconut.constants import (
    wildcard,
    openindent,
    closeindent,
    function_match_error_var,
    match_set_name_var,
    is_data_var,
    default_matcher_style,
    self_match_types,
)
from coconut.compiler.util import (
    paren_join,
    handle_indentation,
    add_int_and_strs,
    ordered_items,
)

# -----------------------------------------------------------------------------------------------------------------------
# UTILITIES:
# -----------------------------------------------------------------------------------------------------------------------


def get_match_names(match):
    """Gets keyword names for the given match."""
    internal_assert(not isinstance(match, str), "invalid match in get_match_names", match)
    names = []
    # these constructs directly contain top-level variable names
    if "var" in match:
        (setvar,) = match
        if setvar != wildcard:
            names.append(setvar)
    elif "as" in match:
        as_match, as_names = match[0], match[1:]
        names.extend(as_names)
        names += get_match_names(as_match)
    # these constructs continue matching on the entire original item,
    #  meaning they can also contain top-level variable names
    elif "paren" in match:
        (paren_match,) = match
        names += get_match_names(paren_match)
    elif "and" in match:
        for and_match in match:
            names += get_match_names(and_match)
    elif "infix" in match:
        infix_match = match[0]
        names += get_match_names(infix_match)
    elif "isinstance_is" in match:
        isinstance_is_match = match[0]
        names += get_match_names(isinstance_is_match)
    elif "class" in match or "data_or_class" in match:
        cls_name, class_matches = match
        if cls_name in self_match_types and len(class_matches) == 1 and len(class_matches[0]) == 1:
            names += get_match_names(class_matches[0][0])
    return names


# -----------------------------------------------------------------------------------------------------------------------
# MATCHER:
# -----------------------------------------------------------------------------------------------------------------------


class Matcher(object):
    """Pattern-matching processor."""
    __slots__ = (
        "comp",
        "original",
        "loc",
        "check_var",
        "style",
        "position",
        "checkdefs",
        "names",
        "name_list",
        "child_groups",
        "guards",
        "parent_names",
    )
    matchers = {
        "dict": lambda self: self.match_dict,
        "sequence": lambda self: self.match_sequence,
        "implicit_tuple": lambda self: self.match_implicit_tuple,
        "lazy": lambda self: self.match_lazy,
        "iter": lambda self: self.match_iter,
        "string_sequence": lambda self: self.match_string_sequence,
        "star": lambda self: self.match_star,
        "const": lambda self: self.match_const,
        "is": lambda self: self.match_is,
        "in": lambda self: self.match_in,
        "var": lambda self: self.match_var,
        "set": lambda self: self.match_set,
        "data": lambda self: self.match_data,
        "class": lambda self: self.match_class,
        "data_or_class": lambda self: self.match_data_or_class,
        "paren": lambda self: self.match_paren,
        "as": lambda self: self.match_as,
        "and": lambda self: self.match_and,
        "or": lambda self: self.match_or,
        "view": lambda self: self.match_view,
        "infix": lambda self: self.match_infix,
        "isinstance_is": lambda self: self.match_isinstance_is,
    }
    valid_styles = (
        "coconut",
        "python",
        "coconut warn",
        "python warn",
        "coconut warn on strict",
        "python warn on strict",
    )

    def __init__(self, comp, original, loc, check_var, style=default_matcher_style, name_list=None, parent_names={}):
        """Creates the matcher."""
        self.comp = comp
        self.original = original
        self.loc = loc
        self.check_var = check_var
        internal_assert(style in self.valid_styles, "invalid Matcher style", style)
        self.style = style
        self.name_list = name_list
        self.position = 0
        self.checkdefs = []
        self.parent_names = parent_names
        self.names = OrderedDict()  # ensures deterministic ordering of name setting code
        self.guards = []
        self.child_groups = []
        self.increment()

    def make_child(self):
        """Get an unregistered child matcher object."""
        return Matcher(self.comp, self.original, self.loc, self.check_var, self.style, self.name_list, self.names)

    def branches(self, num_branches):
        """Create num_branches child matchers, one of which must match for the parent match to succeed."""
        child_group = []
        for _ in range(num_branches):
            new_matcher = self.make_child()
            child_group.append(new_matcher)

        self.child_groups.append(child_group)
        return child_group

    def parameterized_branch(self, parameterization):
        """Create a pseudo-child-group parameterized by `for <parameterization>:`."""
        parameterized_child = self.make_child()
        self.child_groups.append((parameterization, parameterized_child))
        return parameterized_child

    def get_checks(self, position=None):
        """Gets the checks at the position."""
        if position is None:
            position = self.position
        return self.checkdefs[position][0]

    def set_checks(self, checks, position=None):
        """Sets the checks at the position."""
        if position is None:
            position = self.position
        self.checkdefs[position][0] = checks

    checks = property(get_checks, set_checks)

    def get_defs(self, position=None):
        """Gets the defs at the position."""
        if position is None:
            position = self.position
        return self.checkdefs[position][1]

    def set_defs(self, defs, position=None):
        """Sets the defs at the position."""
        if position is None:
            position = self.position
        self.checkdefs[position][1] = defs

    defs = property(get_defs, set_defs)

    def add_check(self, check_item):
        """Adds a check universally."""
        self.checks.append(check_item)

    def add_def(self, def_item):
        """Adds a def universally."""
        self.defs.append(def_item)

    def insert_check(self, index, check_item):
        """Inserts a check universally."""
        self.checks.insert(index, check_item)

    def insert_def(self, index, def_item):
        """Inserts a def universally."""
        self.defs.insert(index, def_item)

    @property
    def using_python_rules(self):
        """Whether the current style uses PEP 622 rules."""
        return self.style.startswith("python")

    def rule_conflict_warn(self, message, if_coconut=None, if_python=None, extra=None):
        """Warns on conflicting style rules if callback was given."""
        if self.style.endswith("warn") or self.style.endswith("strict") and self.comp.strict:
            full_msg = message
            if if_python or if_coconut:
                full_msg += " (" + (if_python if self.using_python_rules else if_coconut) + ")"
            if extra:
                full_msg += " (" + extra + ")"
            if self.style.endswith("strict"):
                full_msg += " (remove --strict to dismiss)"
            logger.warn_err(self.comp.make_err(CoconutSyntaxWarning, full_msg, self.original, self.loc))

    def add_guard(self, cond):
        """Adds cond as a guard."""
        self.guards.append(cond)

    def set_position(self, position):
        """Sets the if-statement position."""
        if position < 0:
            position += len(self.checkdefs)
        while position >= len(self.checkdefs):
            self.checkdefs.append(([], []))
        self.position = position

    def increment(self, by=1):
        """Advances the if-statement position."""
        new_pos = self.position + by
        internal_assert(new_pos > 0, "invalid increment/decrement call to set pos to", new_pos)
        self.set_position(new_pos)

    def decrement(self, by=1):
        """Decrements the if-statement position."""
        self.increment(-by)

    @contextmanager
    def down_a_level(self, by=1):
        """Increment then decrement."""
        self.increment(by)
        try:
            yield
        finally:
            self.decrement(by)

    @contextmanager
    def down_to(self, pos):
        """Increment down to pos."""
        orig_pos = self.position
        self.set_position(max(orig_pos, pos))
        try:
            yield
        finally:
            self.set_position(orig_pos)

    @contextmanager
    def down_to_end(self):
        """Increment down until a new set of checkdefs is reached."""
        orig_pos = self.position
        self.set_position(len(self.checkdefs))
        try:
            yield
        finally:
            self.set_position(orig_pos)

    def get_temp_var(self):
        """Gets the next match_temp var."""
        return self.comp.get_temp_var("match_temp")

    def get_set_name_var(self, name):
        """Gets the var for checking whether a name should be set."""
        return match_set_name_var + "_" + name

    def register_name(self, name):
        """Register a new name at the current position."""
        internal_assert(lambda: name not in self.parent_names and name not in self.names, "attempt to register duplicate name", name)
        self.names[name] = self.position
        if self.name_list is not None and name not in self.name_list:
            self.name_list.append(name)

    def match_var(self, tokens, item, bind_wildcard=False):
        """Matches a variable."""
        varname, = tokens
        if varname == wildcard and not bind_wildcard:
            return
        set_name_var = self.get_set_name_var(varname)
        if varname in self.parent_names:
            # no need to increment if it's from the parent
            self.add_check(set_name_var + " == " + item)
        elif varname in self.names:
            var_pos = self.names[varname]
            with self.down_to(var_pos):
                self.add_check(set_name_var + " == " + item)
        else:
            self.add_def(set_name_var + " = " + item)
            with self.down_a_level():
                self.register_name(varname)

    def match_all_in(self, matches, item):
        """Matches all matches to elements of item."""
        for i, match in enumerate(matches):
            self.match(match, item + "[" + str(i) + "]")

    def check_len_in(self, min_len, max_len, item):
        """Checks that the length of item is in range(min_len, max_len+1)."""
        if max_len is None:
            if min_len:
                self.add_check("_coconut.len(" + item + ") >= " + str(min_len))
        elif min_len == max_len:
            self.add_check("_coconut.len(" + item + ") == " + str(min_len))
        elif not min_len:
            self.add_check("_coconut.len(" + item + ") <= " + str(max_len))
        else:
            self.add_check(str(min_len) + " <= _coconut.len(" + item + ") <= " + str(max_len))

    def match_function(self, args, kwargs, pos_only_match_args=(), match_args=(), star_arg=None, kwd_only_match_args=(), dubstar_arg=None):
        """Matches a pattern-matching function."""
        # before everything, pop the FunctionMatchError from context
        self.add_def(function_match_error_var + " = _coconut_get_function_match_error()")
        with self.down_a_level():

            self.match_in_args_kwargs(pos_only_match_args, match_args, args, kwargs, allow_star_args=star_arg is not None)

            if star_arg is not None:
                self.match(star_arg, args + "[" + str(len(match_args)) + ":]")

            self.match_in_kwargs(kwd_only_match_args, kwargs)

            # go down to end to ensure that all popping from kwargs has been done
            with self.down_to_end():
                if dubstar_arg is None:
                    self.add_check("not " + kwargs)
                else:
                    self.match(dubstar_arg, kwargs)

    def match_in_args_kwargs(self, pos_only_match_args, match_args, args, kwargs, allow_star_args=False):
        """Matches against args or kwargs."""
        req_len = 0
        arg_checks = {}
        # go down a level to ensure we're after the length-checking we do later on
        with self.down_a_level():
            for i, arg in enumerate(pos_only_match_args + match_args):
                if isinstance(arg, tuple):
                    (match, default) = arg
                else:
                    match, default = arg, None
                if i < len(pos_only_match_args):  # faster if arg in pos_only_match_args
                    names = None
                else:
                    names = get_match_names(match)
                if default is None:
                    if not names:
                        req_len = i + 1
                        self.match(match, args + "[" + str(i) + "]")
                    else:
                        arg_checks[i] = (
                            # if i < req_len
                            " and ".join('"' + name + '" not in ' + kwargs for name in names),
                            # if i >= req_len
                            "_coconut.sum((_coconut.len(" + args + ") > " + str(i) + ", "
                            + ", ".join('"' + name + '" in ' + kwargs for name in names)
                            + ")) == 1",
                        )
                        tempvar = self.get_temp_var()
                        self.add_def(
                            tempvar + " = "
                            + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else "
                            + "".join(
                                kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
                                for name in names[:-1]
                            )
                            + kwargs + '.pop("' + names[-1] + '")',
                        )
                        with self.down_a_level():
                            self.match(match, tempvar)
                else:
                    if not names:
                        tempvar = self.get_temp_var()
                        self.add_def(tempvar + " = " + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else " + default)
                        with self.down_a_level():
                            self.match(match, tempvar)
                    else:
                        arg_checks[i] = (
                            # if i < req_len
                            None,
                            # if i >= req_len
                            "_coconut.sum((_coconut.len(" + args + ") > " + str(i) + ", "
                            + ", ".join('"' + name + '" in ' + kwargs for name in names)
                            + ")) <= 1",
                        )
                        tempvar = self.get_temp_var()
                        self.add_def(
                            tempvar + " = "
                            + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else "
                            + "".join(
                                kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
                                for name in names
                            )
                            + default,
                        )
                        with self.down_a_level():
                            self.match(match, tempvar)

        # length checking
        max_len = None if allow_star_args else len(pos_only_match_args) + len(match_args)
        self.check_len_in(req_len, max_len, args)
        for i, (lt_check, ge_check) in ordered_items(arg_checks):
            if i < req_len:
                if lt_check is not None:
                    self.add_check(lt_check)
            else:
                if ge_check is not None:
                    self.add_check(ge_check)

    def match_in_kwargs(self, match_args, kwargs):
        """Matches against kwargs."""
        for match, default in match_args:
            names = get_match_names(match)
            if not names:
                raise CoconutDeferredSyntaxError("keyword-only pattern-matching function arguments must be named", self.loc)
            tempvar = self.get_temp_var()
            self.add_def(
                tempvar + " = "
                + "".join(
                    kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
                    for name in names
                )
                + (default if default is not None else "_coconut_sentinel"),
            )
            with self.down_a_level():
                if default is None:
                    self.add_check(tempvar + " is not _coconut_sentinel")
                self.match(match, tempvar)

    def match_dict(self, tokens, item):
        """Matches a dictionary."""
        if len(tokens) == 1:
            matches, rest = tokens[0], None
        else:
            matches, rest = tokens

        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Mapping)")

        if rest is None:
            self.rule_conflict_warn(
                "found pattern with new behavior in Coconut v2; dict patterns now allow the dictionary being matched against to contain extra keys",
                extra="use explicit '{..., **_}' or '{..., **{}}' syntax to resolve",
            )
            check_len = not self.using_python_rules
        elif rest == "{}":
            check_len = True
            rest = None
        else:
            check_len = False

        if check_len:
            self.add_check("_coconut.len(" + item + ") == " + str(len(matches)))

        seen_keys = set()
        for k, v in matches:
            if k in seen_keys:
                raise CoconutDeferredSyntaxError("duplicate key {k!r} in dictionary pattern".format(k=k), self.loc)
            seen_keys.add(k)
            key_var = self.get_temp_var()
            self.add_def(key_var + " = " + item + ".get(" + k + ", _coconut_sentinel)")
            with self.down_a_level():
                self.add_check(key_var + " is not _coconut_sentinel")
                self.match(v, key_var)

        if rest is not None and rest != wildcard:
            match_keys = [k for k, v in matches]
            rest_item = (
                "dict((k, v) for k, v in "
                + item + ".items() if k not in set(("
                + ", ".join(match_keys) + ("," if len(match_keys) == 1 else "")
                + ")))"
            )
            with self.down_a_level():
                self.match_var([rest], rest_item)

    def proc_sequence_match(self, tokens, iter_match=False):
        """Processes sequence match tokens."""
        seq_groups = []
        seq_type = None
        for group in tokens:
            if "capture" in group:
                group_type = "capture"
                if len(group) > 1:
                    raise CoconutDeferredSyntaxError("sequence/iterable patterns cannot contain multiple consecutive arbitrary-length captures", self.loc)
                group_contents = group[0]
            elif "literal" in group:
                group_type = "elem_matches"
                group_contents = []
                for seq_literal in group:
                    got_seq_type, matches = seq_literal
                    if not iter_match:
                        if seq_type is None:
                            seq_type = got_seq_type
                        elif got_seq_type != seq_type:
                            raise CoconutDeferredSyntaxError("list literals and tuple literals cannot be mixed in sequence patterns", self.loc)
                    group_contents.extend(matches)
            elif "elem" in group:
                group_type = "elem_matches"
                group_contents = group
            elif "string" in group:
                group_type = "string"
                for str_literal in group:
                    if str_literal.startswith("b"):
                        got_seq_type = 'b"'
                    else:
                        got_seq_type = '"'
                    if seq_type is None:
                        seq_type = got_seq_type
                    elif got_seq_type != seq_type:
                        raise CoconutDeferredSyntaxError("string literals and byte literals cannot be mixed in string patterns", self.loc)
                if len(group) == 1:
                    str_item = group[0]
                else:
                    str_item = self.comp.eval_now(" ".join(group))
                group_contents = (str_item, len(self.comp.literal_eval(str_item)))
            elif "f_string" in group:
                group_type = "f_string"
                # f strings are always unicode
                if seq_type is None:
                    seq_type = '"'
                elif seq_type != '"':
                    raise CoconutDeferredSyntaxError("string literals and byte literals cannot be mixed in string patterns", self.loc)
                internal_assert(len(group) == 1, "invalid f string sequence match group", group)
                str_item = group[0]
                group_contents = (str_item, "_coconut.len(" + str_item + ")")
            else:
                raise CoconutInternalException("invalid sequence match group", group)
            seq_groups.append((group_type, group_contents))
        return seq_type, seq_groups

    def handle_sequence(self, seq_type, seq_groups, item, iter_match=False):
        """Handle a processed sequence match."""
        # length check
        if not iter_match:
            min_len_int = 0
            min_len_strs = []
            bounded = True
            for gtype, gcontents in seq_groups:
                if gtype == "capture":
                    bounded = False
                elif gtype == "elem_matches":
                    min_len_int += len(gcontents)
                elif gtype == "string":
                    str_item, str_len = gcontents
                    min_len_int += str_len
                elif gtype == "f_string":
                    str_item, str_len = gcontents
                    min_len_strs.append(str_len)
                else:
                    raise CoconutInternalException("invalid sequence match group type", gtype)
            min_len = add_int_and_strs(min_len_int, min_len_strs)
            max_len = min_len if bounded else None
            self.check_len_in(min_len, max_len, item)

        # match head
        start_ind_int = 0
        start_ind_strs = []
        iterable_var = None
        if seq_groups[0][0] == "elem_matches":
            _, matches = seq_groups.pop(0)
            if not iter_match:
                self.match_all_in(matches, item)
            elif matches:
                iterable_var = self.get_temp_var()
                self.add_def(iterable_var + " = _coconut.iter(" + item + ")")
                head_var = self.get_temp_var()
                self.add_def(head_var + " = _coconut.tuple(_coconut_iter_getitem(" + iterable_var + ", _coconut.slice(None, " + str(len(matches)) + ")))")
                with self.down_a_level():
                    self.add_check("_coconut.len(" + head_var + ") == " + str(len(matches)))
                    self.match_all_in(matches, head_var)
            start_ind_int += len(matches)
        elif seq_groups[0][0] == "string":
            internal_assert(not iter_match, "cannot be both string and iter match")
            _, (str_item, str_len) = seq_groups.pop(0)
            if str_len > 0:
                self.add_check(item + ".startswith(" + str_item + ")")
            start_ind_int += str_len
        elif seq_groups[0][0] == "f_string":
            internal_assert(not iter_match, "cannot be both f string and iter match")
            _, (str_item, str_len) = seq_groups.pop(0)
            self.add_check(item + ".startswith(" + str_item + ")")
            start_ind_strs.append(str_len)
        if not seq_groups:
            return
        start_ind = add_int_and_strs(start_ind_int, start_ind_strs)

        # match tail
        last_ind_int = -1
        last_ind_strs = []
        if seq_groups[-1][0] == "elem_matches":
            internal_assert(not iter_match, "iter_match=True should not be passed for tail patterns")
            _, matches = seq_groups.pop()
            for i, match in enumerate(matches):
                self.match(match, item + "[-" + str(len(matches) - i) + "]")
            last_ind_int -= len(matches)
        elif seq_groups[-1][0] == "string":
            internal_assert(not iter_match, "cannot be both string and iter match")
            _, (str_item, str_len) = seq_groups.pop()
            if str_len > 0:
                self.add_check(item + ".endswith(" + str_item + ")")
            last_ind_int -= str_len
        elif seq_groups[-1][0] == "f_string":
            internal_assert(not iter_match, "cannot be both f string and iter match")
            _, (str_item, str_len) = seq_groups.pop()
            self.add_check(item + ".endswith(" + str_item + ")")
            last_ind_strs.append("-" + str_len)
        if not seq_groups:
            return
        last_ind = add_int_and_strs(last_ind_int, last_ind_strs)

        # we need to go down a level to ensure we're below 'match head' above
        with self.down_a_level():

            # extract middle
            cache_mid_item = False
            if iterable_var is None:

                # make middle by indexing into item
                start_ind_str = "" if start_ind == 0 else str(start_ind)
                last_ind_str = "" if last_ind == -1 else str(last_ind + 1) if isinstance(last_ind, int) else last_ind + " + 1"
                if start_ind_str or last_ind_str:
                    mid_item = item + "[" + start_ind_str + ":" + last_ind_str + "]"
                    cache_mid_item = True
                else:
                    mid_item = item

                # convert middle to proper sequence type if necessary
                if seq_type == "[":
                    mid_item = "_coconut.list(" + mid_item + ")"
                    cache_mid_item = True
                elif seq_type == "(":
                    mid_item = "_coconut.tuple(" + mid_item + ")"
                    cache_mid_item = True
                elif seq_type in (None, '"', 'b"', "(|"):
                    # if we know mid_item is already the desired type, no conversion is needed
                    pass
                elif seq_type is False:
                    raise CoconutInternalException("attempted to convert with to_sequence when seq_type was marked as False", mid_item)
                else:
                    raise CoconutInternalException("invalid sequence match type", seq_type)

            else:
                mid_item = iterable_var

            # cache middle
            if cache_mid_item:
                cached_mid_item = self.get_temp_var()
                self.add_def(cached_mid_item + " = " + mid_item)
                mid_item = cached_mid_item

            # we need to go down a level to ensure we're below 'cache middle' above
            with self.down_a_level():

                # handle single-capture middle
                if len(seq_groups) == 1:
                    gtype, match = seq_groups[0]
                    internal_assert(gtype == "capture", "invalid sequence match middle group", seq_groups)
                    self.match(match, mid_item)
                    return
                internal_assert(len(seq_groups) >= 3, "invalid sequence match middle groups", seq_groups)

                # handle linear search patterns
                if len(seq_groups) == 3:
                    (front_gtype, front_match), mid_group, (back_gtype, back_match) = seq_groups
                    internal_assert(front_gtype == "capture" == back_gtype, "invalid sequence match middle groups", seq_groups)
                    mid_gtype, mid_contents = mid_group

                    if iter_match:
                        internal_assert(mid_gtype == "elem_matches", "invalid iterable search match middle group", mid_group)
                        mid_len = len(mid_contents)
                        if mid_len == 0:
                            raise CoconutDeferredSyntaxError("found empty iterable search pattern", self.loc)

                        # ensure we have an iterable var to work with
                        if iterable_var is None:
                            iterable_var = self.get_temp_var()
                            self.add_def(iterable_var + " = _coconut.iter(" + mid_item + ")")

                        # create a cache variable to store elements so far
                        iter_cache_var = self.get_temp_var()
                        self.add_def(iter_cache_var + " = []")

                        # construct a parameterized child to perform the search
                        iter_item_var = self.get_temp_var()
                        parameterized_child = self.parameterized_branch(
                            "for {iter_item_var} in {iterable_var}".format(
                                iter_item_var=iter_item_var,
                                iterable_var=iterable_var,
                            ),
                        )
                        parameterized_child.add_def(iter_cache_var + ".append(" + iter_item_var + ")")
                        with parameterized_child.down_a_level():
                            parameterized_child.add_check("_coconut.len(" + iter_cache_var + ") >= " + str(mid_len))

                            # get the items to search against
                            front_item = "{iter_cache_var}[:-{mid_len}]".format(iter_cache_var=iter_cache_var, mid_len=mid_len)
                            searching_through = "{iter_cache_var}[-{mid_len}:]".format(iter_cache_var=iter_cache_var, mid_len=mid_len)
                            back_item = iterable_var

                            # perform the matches in the child
                            search_item = parameterized_child.get_temp_var()
                            parameterized_child.add_def(search_item + " = " + searching_through)
                            with parameterized_child.down_a_level():
                                parameterized_child.handle_sequence(seq_type, [mid_group], search_item)
                                # no need to make temp_vars here since these are guaranteed to be var matches
                                parameterized_child.match(front_match, front_item)
                                parameterized_child.match(back_match, back_item)

                    elif mid_gtype == "elem_matches":
                        mid_len = len(mid_contents)

                        # construct a parameterized child to perform the search
                        seq_ind_var = self.get_temp_var()
                        parameterized_child = self.parameterized_branch(
                            "for {seq_ind_var} in _coconut.range(_coconut.len({mid_item}))".format(
                                seq_ind_var=seq_ind_var,
                                mid_item=mid_item,
                            ),
                        )

                        # get the items to search against
                        front_item = "{mid_item}[:{seq_ind_var}]".format(
                            mid_item=mid_item,
                            seq_ind_var=seq_ind_var,
                        )
                        searching_through = "{mid_item}[{seq_ind_var}:{seq_ind_var} + {mid_len}]".format(
                            mid_item=mid_item,
                            seq_ind_var=seq_ind_var,
                            mid_len=mid_len,
                        )
                        back_item = "{mid_item}[{seq_ind_var} + {mid_len}:]".format(
                            mid_item=mid_item,
                            seq_ind_var=seq_ind_var,
                            mid_len=mid_len,
                        )

                        # perform the matches in the child
                        search_item = parameterized_child.get_temp_var()
                        parameterized_child.add_def(search_item + " = " + searching_through)
                        with parameterized_child.down_a_level():
                            parameterized_child.handle_sequence(seq_type, [mid_group], search_item)
                            # these are almost always var matches, so we don't bother to make new temp vars here
                            parameterized_child.match(front_match, front_item)
                            parameterized_child.match(back_match, back_item)

                    elif mid_gtype in ("string", "f_string"):
                        str_item, str_len = mid_contents
                        found_loc = self.get_temp_var()
                        self.add_def(found_loc + " = " + mid_item + ".find(" + str_item + ")")
                        with self.down_a_level():
                            self.add_check(found_loc + " != -1")
                            # no need to make temp_vars here since these are guaranteed to be var matches
                            self.match(front_match, "{mid_item}[:{found_loc}]".format(mid_item=mid_item, found_loc=found_loc))
                            self.match(back_match, "{mid_item}[{found_loc} + {str_len}:]".format(mid_item=mid_item, found_loc=found_loc, str_len=str_len))

                    else:
                        raise CoconutInternalException("invalid linear search group type", mid_gtype)
                    return

                # raise on unsupported quadratic matches
                raise CoconutDeferredSyntaxError("nonlinear sequence search patterns are not supported", self.loc)

    def match_sequence(self, tokens, item):
        """Matches an arbitrary sequence pattern."""
        internal_assert(len(tokens) >= 1, "invalid sequence match tokens", tokens)

        # abc check
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Sequence)")

        # extract groups
        seq_type, seq_groups = self.proc_sequence_match(tokens)

        # match sequence
        self.handle_sequence(seq_type, seq_groups, item)

    def match_lazy(self, tokens, item):
        """Matches lazy lists."""
        (seq_type, matches), = tokens
        internal_assert(seq_type == "(|", "invalid lazy list match tokens", tokens)

        # abc check
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")

        # match sequence
        temp_item_var = self.get_temp_var()
        self.add_def(temp_item_var + " = _coconut.tuple(" + item + ")")
        with self.down_a_level():
            self.handle_sequence(False, [["elem_matches", matches]], temp_item_var)

    def match_implicit_tuple(self, tokens, item):
        """Matches an implicit tuple."""
        internal_assert(len(tokens) >= 1, "invalid implicit tuple tokens", tokens)

        # abc check
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")

        # match sequence
        temp_item_var = self.get_temp_var()
        self.add_def(temp_item_var + " = _coconut.tuple(" + item + ")")
        with self.down_a_level():
            self.handle_sequence(False, [["elem_matches", tokens]], temp_item_var)

    def match_star(self, tokens, item):
        """Matches starred assignment."""
        internal_assert(len(tokens) >= 1, "invalid star match tokens", tokens)

        # abc check
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")

        # extract groups
        _, seq_groups = self.proc_sequence_match(tokens)

        # match sequence
        temp_item_var = self.get_temp_var()
        self.add_def(temp_item_var + " = _coconut.list(" + item + ")")
        with self.down_a_level():
            self.handle_sequence(None, seq_groups, temp_item_var)

    def match_string_sequence(self, tokens, item):
        """Match string sequence patterns."""
        seq_type, seq_groups = self.proc_sequence_match(tokens)

        # type check
        if seq_type == '"':
            self.add_check("_coconut.isinstance(" + item + ", _coconut.str)")
        elif seq_type == 'b"':
            self.add_check("_coconut.isinstance(" + item + ", _coconut.bytes)")
        else:
            raise CoconutInternalException("invalid string match type", seq_type)

        # match sequence
        self.handle_sequence(seq_type, seq_groups, item)

    def match_iter(self, tokens, item):
        """Matches a chain."""
        internal_assert(len(tokens) >= 1, "invalid iterable match tokens", tokens)

        # abc check
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")

        # match iterable
        _, seq_groups = self.proc_sequence_match(tokens, iter_match=True)
        if seq_groups[-1][0] == "elem_matches":  # tail pattern
            temp_item_var = self.get_temp_var()
            self.add_def(temp_item_var + " = _coconut.tuple(" + item + ")")
            with self.down_a_level():
                self.handle_sequence(None, seq_groups, temp_item_var)
        else:
            self.handle_sequence(None, seq_groups, item, iter_match=True)

    def match_const(self, tokens, item):
        """Matches an equality check."""
        match, = tokens
        self.add_check(item + " == " + match)

    def match_is(self, tokens, item):
        """Matches an identity check."""
        match, = tokens
        self.add_check(item + " is " + match)

    def match_in(self, tokens, item):
        """Matches a containment check."""
        match, = tokens
        self.add_check(item + " in " + match)

    def match_set(self, tokens, item):
        """Matches a set."""
        match, = tokens
        self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Set)")
        self.add_check("_coconut.len(" + item + ") == " + str(len(match)))
        for const in match:
            self.add_check(const + " in " + item)

    def split_data_or_class_match(self, tokens):
        """Split data/class match tokens into cls_name, pos_matches, name_matches, star_match."""
        cls_name, matches = tokens

        pos_matches = []
        name_matches = {}
        star_match = None
        for match_arg in matches:
            # positional arg
            if len(match_arg) == 1:
                match, = match_arg
                if star_match is not None:
                    raise CoconutDeferredSyntaxError("positional arg after starred arg in data/class match", self.loc)
                if name_matches:
                    raise CoconutDeferredSyntaxError("positional arg after keyword arg in data/class match", self.loc)
                pos_matches.append(match)
            # starred arg
            elif len(match_arg) == 2:
                internal_assert(match_arg[0] == "*", "invalid starred data/class match arg tokens", match_arg)
                _, match = match_arg
                if star_match is not None:
                    raise CoconutDeferredSyntaxError("duplicate starred arg in data/class match", self.loc)
                if name_matches:
                    raise CoconutDeferredSyntaxError("both starred arg and keyword arg in data/class match", self.loc)
                star_match = match
            # keyword arg
            else:
                if len(match_arg) == 3:
                    internal_assert(match_arg[1] == "=", "invalid keyword data/class match arg tokens", match_arg)
                    name, _, match = match_arg
                    strict = False
                elif len(match_arg) == 4:
                    internal_assert(match_arg[0] == "." and match_arg[2] == "=", "invalid strict keyword data/class match arg tokens", match_arg)
                    _, name, _, match = match_arg
                    strict = True
                else:
                    raise CoconutInternalException("invalid data/class match arg", match_arg)
                if star_match is not None:
                    raise CoconutDeferredSyntaxError("both keyword arg and starred arg in data/class match", self.loc)
                if name in name_matches:
                    raise CoconutDeferredSyntaxError("duplicate keyword arg {name!r} in data/class match".format(name=name), self.loc)
                name_matches[name] = (match, strict)

        return cls_name, pos_matches, name_matches, star_match

    def match_class_attr(self, match, attr, item):
        """Match an attribute for a class match where attr is an expression that evaluates to the attribute name."""
        attr_var = self.get_temp_var()
        self.add_def(attr_var + " = _coconut.getattr(" + item + ", " + attr + ", _coconut_sentinel)")
        with self.down_a_level():
            self.add_check(attr_var + " is not _coconut_sentinel")
            self.match(match, attr_var)

    def match_class_names(self, name_matches, item):
        """Matches keyword class patterns."""
        for name, (match, strict) in name_matches.items():
            if strict:
                self.match(match, item + "." + name)
            else:
                self.match_class_attr(match, ascii(name), item)

    def match_class(self, tokens, item):
        """Matches a class PEP-622-style."""
        cls_name, pos_matches, name_matches, star_match = self.split_data_or_class_match(tokens)

        self.add_check("_coconut.isinstance(" + item + ", " + cls_name + ")")

        self_match_matcher, other_cls_matcher = self.branches(2)

        # handle instances of _coconut_self_match_types
        self_match_matcher.add_check("_coconut.type(" + item + ") in _coconut_self_match_types")
        if pos_matches:
            if len(pos_matches) > 1:
                self_match_matcher.add_def(
                    handle_indentation(
                        """
raise _coconut.TypeError("too many positional args in class match (pattern requires {num_pos_matches}; '{cls_name}' only supports 1)")
                    """,
                    ).format(
                        num_pos_matches=len(pos_matches),
                        cls_name=cls_name,
                    ),
                )
            else:
                self_match_matcher.match(pos_matches[0], item)

        # handle all other classes
        other_cls_matcher.add_check("not _coconut.type(" + item + ") in _coconut_self_match_types")
        match_args_var = other_cls_matcher.get_temp_var()
        other_cls_matcher.add_def(
            handle_indentation("""
{match_args_var} = _coconut.getattr({cls_name}, '__match_args__', ())
if not _coconut.isinstance({match_args_var}, _coconut.tuple):
    raise _coconut.TypeError("{cls_name}.__match_args__ must be a tuple")
if _coconut.len({match_args_var}) < {num_pos_matches}:
    raise _coconut.TypeError("too many positional args in class match (pattern requires {num_pos_matches}; '{cls_name}' only supports %s)" % (_coconut.len({match_args_var}),))
        """).format(
                cls_name=cls_name,
                match_args_var=match_args_var,
                num_pos_matches=len(pos_matches),
            ),
        )
        with other_cls_matcher.down_a_level():
            for i, match in enumerate(pos_matches):
                other_cls_matcher.match_class_attr(match, match_args_var + "[" + str(i) + "]", item)

        # handle starred arg
        if star_match is not None:
            star_match_var = self.get_temp_var()
            self.add_def(
                handle_indentation(
                    """
{match_args_var} = _coconut.getattr({cls_name}, '__match_args__', ())
{star_match_var} = _coconut.tuple(_coconut.getattr({item}, {match_args_var}[i]) for i in _coconut.range({num_pos_matches}, _coconut.len({match_args_var})))
                """,
                ).format(
                    match_args_var=self.get_temp_var(),
                    cls_name=cls_name,
                    star_match_var=star_match_var,
                    item=item,
                    num_pos_matches=len(pos_matches),
                ),
            )
            with self.down_a_level():
                self.match(star_match, star_match_var)

        # handle keyword args
        self.match_class_names(name_matches, item)

    def match_data(self, tokens, item):
        """Matches a data type."""
        cls_name, pos_matches, name_matches, star_match = self.split_data_or_class_match(tokens)

        self.add_check("_coconut.isinstance(" + item + ", " + cls_name + ")")

        if star_match is None:
            self.add_check(
                '_coconut.len({item}) == {total_len}'.format(
                    item=item,
                    total_len=len(pos_matches) + len(name_matches),
                ),
            )
        # avoid checking >= 0
        elif len(pos_matches):
            self.add_check(
                "_coconut.len({item}) >= {min_len}".format(
                    item=item,
                    min_len=len(pos_matches),
                ),
            )

        self.match_all_in(pos_matches, item)

        if star_match is not None:
            self.match(star_match, item + "[" + str(len(pos_matches)) + ":]")

        # handle keyword args
        self.match_class_names(name_matches, item)

    def match_data_or_class(self, tokens, item):
        """Matches an ambiguous data or class match."""
        cls_name, matches = tokens

        is_data_result_var = self.get_temp_var()
        self.add_def(
            handle_indentation(
                """
{is_data_result_var} = _coconut.getattr({cls_name}, "{is_data_var}", False) or _coconut.isinstance({cls_name}, _coconut.tuple) and _coconut.all(_coconut.getattr(_coconut_x, "{is_data_var}", False) for _coconut_x in {cls_name}){type_comment}
            """,
            ).format(
                is_data_result_var=is_data_result_var,
                is_data_var=is_data_var,
                cls_name=cls_name,
                type_comment=self.comp.type_ignore_comment(),
            ),
        )

        if_data, if_class = self.branches(2)

        if_data.add_check(is_data_result_var)
        if_data.match_data(tokens, item)

        if_class.add_check("not " + is_data_result_var)
        if_class.match_class(tokens, item)

    def match_paren(self, tokens, item):
        """Matches a paren."""
        match, = tokens
        return self.match(match, item)

    def match_as(self, tokens, item):
        """Matches as patterns."""
        match, as_names = tokens[0], tokens[1:]
        for varname in as_names:
            self.match_var([varname], item, bind_wildcard=True)
        self.match(match, item)

    def match_isinstance_is(self, tokens, item):
        """Matches old-style isinstance checks."""
        match, isinstance_checks = tokens[0], tokens[1:]

        if "var" in match:
            varname, = match
        else:
            varname = "..."
        isinstance_checks_str = varname + " is " + " is ".join(isinstance_checks)
        cls_syntax = " and ".join(instcheck + "(" + varname + ")" for instcheck in isinstance_checks)
        explicit_syntax = " and ".join(varname + " `isinstance` " + instcheck for instcheck in isinstance_checks)
        self.comp.strict_err_or_warn(
            "found deprecated isinstance-checking " + repr(isinstance_checks_str) + " pattern;"
            " rewrite to use class patterns (try " + repr(cls_syntax) + ") or explicit isinstance-checking (" + repr(explicit_syntax) + " should always work)",
            self.original,
            self.loc,
        )

        for instcheck in isinstance_checks:
            self.add_check("_coconut.isinstance(" + item + ", " + instcheck + ")")
        self.match(match, item)

    def match_and(self, tokens, item):
        """Matches and."""
        for match in tokens:
            self.match(match, item)

    def match_or(self, tokens, item):
        """Matches or."""
        new_matchers = self.branches(len(tokens))
        for m, tok in zip(new_matchers, tokens):
            m.match(tok, item)

    def match_view(self, tokens, item):
        """Matches view patterns"""
        view_func, view_pattern = tokens

        func_result_var = self.get_temp_var()
        self.add_def(
            handle_indentation(
                """
try:
    {func_result_var} = ({view_func})({item})
except _coconut.Exception as _coconut_view_func_exc:
    if _coconut.getattr(_coconut_view_func_exc.__class__, "__name__", None) == "MatchError":
        {func_result_var} = _coconut_sentinel
    else:
        raise
            """,
            ).format(
                func_result_var=func_result_var,
                view_func=view_func,
                item=item,
            ),
        )

        with self.down_a_level():
            self.add_check(func_result_var + " is not _coconut_sentinel")
            self.match(view_pattern, func_result_var)

    def match_infix(self, tokens, item):
        """Matches infix patterns."""
        match, infix_toks = tokens[0], tokens[1:]

        for toks in infix_toks:
            if len(toks) == 1:
                op, arg = toks[0], None
            else:
                op, arg = toks

            infix_check = "(" + op + ")(" + item
            if arg is not None:
                infix_check += ", " + arg
            infix_check += ")"
            self.add_check(infix_check)

        self.match(match, item)

    def make_match(self, flag, tokens):
        """Create an artificial match object."""
        return ParseResults(tokens, flag)

    def match(self, tokens, item):
        """Performs pattern-matching processing."""
        internal_assert(not isinstance(tokens, str), "invalid match tokens", tokens)
        for flag, get_handler in self.matchers.items():
            if flag in tokens:
                return get_handler(self)(tokens, item)
        raise CoconutInternalException("invalid pattern-matching tokens", tokens)

    def out(self):
        """Return pattern-matching code assuming check_var starts False."""
        out = []

        # set match_set_name_vars to sentinels
        for name in self.names:
            out.append(self.get_set_name_var(name) + " = _coconut_sentinel\n")

        # match checkdefs setting check_var
        closes = 0
        for checks, defs in self.checkdefs:
            if checks:
                out.append("if " + paren_join(checks, "and") + ":\n" + openindent)
                closes += 1
            if defs:
                out.append("\n".join(defs) + "\n")
        out.append(self.check_var + " = True\n" + closeindent * closes)

        # handle children
        for children in self.child_groups:
            internal_assert(children, "got child group with no children", self.child_groups)

            # handle parameterized child groups
            if isinstance(children[0], str):
                parameterization, child = children
                out.append(
                    handle_indentation(
                        """
if {check_var}:
    {check_var} = False
    {parameterization}:
        {child_checks}
        if {check_var}:
            break
                        """,
                        add_newline=True,
                    ).format(
                        check_var=self.check_var,
                        parameterization=parameterization,
                        child_checks=child.out().rstrip(),
                    ),
                )

            # handle normal child groups
            else:
                children_checks = "".join(
                    handle_indentation(
                        """
if not {check_var}:
    {child_out}
                        """,
                        add_newline=True,
                    ).format(
                        check_var=self.check_var,
                        child_out=child.out(),
                    ) for child in children
                )
                out.append(
                    handle_indentation(
                        """
if {check_var}:
    {check_var} = False
    {children_checks}
                        """,
                        add_newline=True,
                    ).format(
                        check_var=self.check_var,
                        children_checks=children_checks,
                    ),
                )

        # commit variable definitions
        name_set_code = []
        for name in self.names:
            name_set_code.append(
                handle_indentation(
                    """
if {set_name_var} is not _coconut_sentinel:
    {name} = {set_name_var}
                    """,
                    add_newline=True,
                ).format(
                    set_name_var=self.get_set_name_var(name),
                    name=name,
                ),
            )
        if name_set_code:
            out.append(
                handle_indentation(
                    """
if {check_var}:
    {name_set_code}
                    """,
                ).format(
                    check_var=self.check_var,
                    name_set_code="".join(name_set_code),
                ),
            )

        # handle guards
        if self.guards:
            out.append(
                handle_indentation(
                    """
if {check_var} and not ({guards}):
    {check_var} = False
                    """,
                    add_newline=True,
                ).format(
                    check_var=self.check_var,
                    guards=paren_join(self.guards, "and"),
                ),
            )

        return "".join(out)

    def build(self, stmts=None, set_check_var=True, invert=False):
        """Construct code for performing the match then executing stmts."""
        out = []
        if set_check_var:
            out.append(self.check_var + " = False\n")
        out.append(self.out())
        if stmts is not None:
            out.append("if " + ("not " if invert else "") + self.check_var + ":" + "\n" + openindent + "".join(stmts) + closeindent)
        return "".join(out)
